#!/usr/bin/env python
# coding: utf-8

# # PRIMARY FILTER FUNCTION
# ### Uses base functions to compute affordability metrics.


from warnings import filterwarnings

import pandas as pd

from mortgage_filter.exceptions import *
from mortgage_filter.mortgage_base import *


def property_filter(
    property_data,
    downpayment,
    mortgage_rate=None,
    mortgage_term=None,
    max_monthly_payment=None,
    max_loan=None,
):
    """Given a dataframe of properties, their prices, and some basic financial information, it returns a dataframe with only the affordable properties and other affordability metrics (ie. how long it would take to pay off, monthly payments, total interest, etc.).

    Arguments
    ----------
    data : dataframe
        Areas/properties in column index 0 (str)
        Respective prices in column index 1 (numeric)

    downpayment : numeric
        Your maximal possible downpayment

    mortgage_rate : numeric
        Interest rate on the mortgage loan (leave empty if mortgage_term is provided)

    mortgage_term : int
        Contract length in years (1 to 10) for the mortgage interest rate.
        Only specify if you do not know what mortgage_rate to enter (leave empty if mortgage_rate provided)

    max_monthly_payment : numeric
        Your max affordable or bank limited monthly payment towards your home

    max_loan : numeric
        Max eligible loan based on your downpayment

    Return
    ------
    dataframe
        Properties/Areas
        Prices/Average area price
        Minimum_Downpayment
        Mortgage_Insurance
        Principal
        Monthly_Payment
        Shortest_Amortization
        Total_Interest
        Net_Cost (assuming no other fees)
    """

    filterwarnings("ignore")

    try:
        # is object a dataframe?
        if isinstance(property_data, pd.DataFrame) == False:
            raise FilterInputError("Dataframe object expected")
        # right number of columns?
        if len(property_data.columns) != 2:
            raise FormatError(
                "Expected two columns of type str and numeric, respectively"
            )
        # is column at index 1 (price) numeric?
        if pd.api.types.is_numeric_dtype(property_data.iloc[:, 1]) == False:
            raise TypeError("Column at index 1 (price) must be numeric")

    except FilterInputError as FIE:
        print(FIE, "\nReveived object of type:", type(property_data))
        return None
    except FormatError as FE:
        print(
            FE,
            "\nReveived dataframe with this many columns:",
            len(property_data.columns),
        )
        return None
    except TypeError as TE:
        print(TE)
        return None

    data = property_data.copy()

    # Rename columns
    data = data.set_axis(["Property/Area", "Price"], axis=1)

    # Note original input of properties
    og_prop_count = data["Property/Area"].count()

    # FILTER: Downpayment. Remove properties where minimal DP exceeds your entered DP
    data["Minimum_Downpayment"] = data.iloc[:, 1].apply(lambda x: min_downpayment(x))
    data = data[data["Minimum_Downpayment"] <= downpayment]

    # Mortgage rate. If none provided give a reasonable estimate
    if mortgage_rate == None:
        mortgage_rate = mort_rate(mortgage_term)

    # Calculate mortgage insurance (default insurance) lump sum for each property
    data["Mortgage_Insurance"] = data.loc[:, "Price"].apply(
        lambda p: mortgage_insurance(p, downpayment)
    )

    # Calculate initial principal for each property
    data["Principal"] = round(
        (data["Price"] - downpayment + data["Mortgage_Insurance"]), 2
    )

    # FILTER: Max eligible loan. Remove properties where the principal exceeds the max approved loan
    # If no max loan specified assume no limit
    if max_loan != None:
        data = data[data["Principal"] < max_loan]

    # Add two columns for monthly payment and shortest amortization period
    # These are outputs of the optimal_monthly_payment function
    temp = data.loc[:, "Principal"].apply(
        lambda principal: optimal_monthly_payment(
            principal, mortgage_rate, max_monthly_payment
        )
    )
    temp = list(zip(*temp))

    data["Monthly_Payment"] = temp[0]
    data["Shortest_Amortization"] = temp[1]

    # FILTER: Remove rows where Monthly_Payment is NaN
    data = data[data["Monthly_Payment"].notnull()]

    # Add column for the cumulative cost of interest given that amortization
    tot_int = []
    for princ, monthly_payment in data[["Principal", "Monthly_Payment"]].itertuples(
        index=False
    ):
        tot_int += [total_interest(princ, mortgage_rate, monthly_payment)]
    data["Total_Interest"] = tot_int

    # Add column for net cost of home (price + cumulative interest + mortgage insurance)
    data["Net_Cost"] = (
        data["Price"] + data["Mortgage_Insurance"] + data["Total_Interest"]
    )

    print(
        f"You can afford {data['Property/Area'].count()} properties from the {og_prop_count} you've provided."
    )
    return data



#!/usr/bin/env python
# coding: utf-8

import unittest

from mortgage_filter.mortgage_filter import *

# from mortgage_package.exceptions import *


class Test_filter(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        ## Import dependencies
        from numpy import linspace
        from pandas import read_csv

        print("!!! Class setup for Test(Test_filter) !!!")

    def setUp(self):
        ## Import property dataset
        self.df = pd.read_csv("data/vancouver_testset.csv", header=0)
        self.df = self.df.loc[:, ["Area", "House Price"]]

        ## Prior knowns for 4 test sets for the property filter
        self.downpayment = np.linspace(100000, 600000, 4)
        self.mortgage_rate = np.linspace(1.6, 3.8, 4)
        self.mortgage_term = np.linspace(1, 10, 4)
        self.max_monthly_payment = np.linspace(1500, 11000, 4)
        self.max_loan = np.linspace(100000, 600000, 4) / 0.05

        print("~~~ start test ~~~")

    def test_property_filter(self):
        self.expected1 = ["East Burnaby", "North Vancouver", "North Vancouver"]
        self.expected2 = ["East Burnaby", "North Vancouver", "West Vancouver"]

        ## Test first case to ensure returned dataframe is empty as expected.
        self.assertTrue(
            property_filter(
                property_data=self.df,
                downpayment=self.downpayment[0],
                mortgage_rate=self.mortgage_rate[0],
                max_monthly_payment=self.max_monthly_payment[0],
                max_loan=self.max_loan[0],
            ).empty
        )
        self.assertTrue(
            property_filter(
                property_data=self.df,
                downpayment=self.downpayment[0],
                mortgage_term=self.mortgage_term[0],
                max_monthly_payment=self.max_monthly_payment[0],
                max_loan=self.max_loan[0],
            ).empty
        )

        ## Test to ensure handling of bad input objects
        ## Dataframe has wrong column type for price
        self.assertIsNone(
            property_filter(
                property_data=self.df.astype("string"),
                downpayment=self.downpayment[0],
                mortgage_term=self.mortgage_term[0],
                max_monthly_payment=self.max_monthly_payment[0],
                max_loan=self.max_loan[0],
            )
        )
        ## Non dataframe object given
        self.assertIsNone(
            property_filter(
                property_data="NOT a dataframe object",
                downpayment=self.downpayment[0],
                mortgage_term=self.mortgage_term[0],
                max_monthly_payment=self.max_monthly_payment[0],
                max_loan=self.max_loan[0],
            )
        )
        ## Dataframe doesnt have 2 columns (must have 2)
        self.x = self.df.drop("Area", axis=1)
        self.assertIsNone(
            property_filter(
                property_data=self.x,
                downpayment=self.downpayment[0],
                mortgage_term=self.mortgage_term[0],
                max_monthly_payment=self.max_monthly_payment[0],
                max_loan=self.max_loan[0],
            )
        )

        ## Test last three cases where return is a non empty dataframe
        for i in range(1, 4):

            ## Testing with known mortgage rate
            self.assertEqual(
                (
                    property_filter(
                        property_data=self.df,
                        downpayment=self.downpayment[i],
                        mortgage_rate=self.mortgage_rate[i],
                        max_monthly_payment=self.max_monthly_payment[i],
                        max_loan=self.max_loan[i],
                    ).iloc[0, 0]
                ),
                self.expected1[i - 1],
            )

            ## Testing with unknown mortgage rate
            self.assertEqual(
                property_filter(
                    property_data=self.df,
                    downpayment=self.downpayment[i],
                    mortgage_term=self.mortgage_term[i],
                    max_monthly_payment=self.max_monthly_payment[i],
                    max_loan=self.max_loan[i],
                ).iloc[0, 0],
                self.expected2[i - 1],
            )

    def tearDown(self):
        print("~~~ complete ~~~")

    @classmethod
    def tearDownClass(cls):
        print("!!! Class teardown for Test(Test_filter) !!!\n\n\n")


# unittest.main(argv=[''], verbosity=2, exit=False)



#!/usr/bin/env python
# coding: utf-8


class mortgageError(Exception):
    """Mortgage subpackage custom exception class."""

    pass


class FilterInputError(mortgageError):
    """Raised when property filter receives a non dataframe object"""

    pass


class FormatError(mortgageError):
    """Raised when a function's input has an improper format"""

    pass


class TermError(mortgageError):
    """Raised if term length is is unavailable. Terms must range from (1,10) years."""

    pass


class MinimalDPError(mortgageError):
    """Raised when downpayment input value is too low to be legally considered"""

    pass



#!/usr/bin/env python
# coding: utf-8

from mortgage_filter.mortgage_base import *

import unittest

class Test_base_functions(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        ## Import dependencies
        from numpy import nan, linspace
        print('!!! Class setup for Test(Test_base_functions) !!!')
    
    def setUp(self):
        ## Test values
        self.downpayment = [0, 0, 0, 200000, 150000, 50000]
        self.max_monthly_payment = np.linspace(1500, 11000, 4)
        self.principal = 600000
        self.mortgage_rate = 2.6
        self.price = ['4000000', -400000, 0, 400000, 800000, 1200000]
        self.term = [-1, 0.5, 1, 5.5, 10, 12]
        self.amortization_years = [1, 8, 16, 24]
        self.amortization_months = [12, 96, 192, 288]
        print('~~~ start test ~~~')
        
        
    def test_min_downpayment(self):
        self.expected = [None, None, 0, 20000, 55000, 240000]
        for i in range(6):
            self.assertEqual(min_downpayment(self.price[i]), self.expected[i])   
        
    def test_mort_rate(self):
        self.expected = [1.736, 1.672, 1.668, 2.005, 3.018, 3.686]
        for i in range(6):
            self.assertEqual(mort_rate(self.term[i]), self.expected[i])
        
    def test_mortgage_insurance(self):
        self.expected = [None, None, None, 0, 16026.85, None]
        for i in range(6):
            self.assertEqual(mortgage_insurance(self.price[i], self.downpayment[i]), self.expected[i])
    
    def test_monthly_payment(self):
        self.expected = [50706.96, 6929.26, 3823.27, 2802.68]
        for i in range(4):
            self.assertEqual(monthly_payment(self.principal, self.mortgage_rate, self.amortization_years[i], months = False),
                             self.expected[i])
            self.assertEqual(monthly_payment(self.principal, self.mortgage_rate, self.amortization_months[i], months = True),
                             self.expected[i])
        
    def test_optimal_monthly_payment(self):
        self.expected = [[np.nan, np.nan], [4536.83, 13], [7820.28, 7], [10674.89, 5]]
        for i in range(4):
            self.assertEqual(optimal_monthly_payment(self.principal, self.mortgage_rate, self.max_monthly_payment[i]),
                             self.expected[i])

    def test_total_interest(self):
        self.expected = [796440.43, 104031.41, 56784.2, 39211.53]
        for i in range(4):
            self.assertEqual(total_interest(self.principal, self.mortgage_rate, self.max_monthly_payment[i]),
                             self.expected[i])
        
        ## handling test for insufficient monthly income leading to infinitely large interest
        self.assertIsNone(total_interest(self.principal, self.mortgage_rate, 100), None)
        
            
    def tearDown(self):
        print('~~~ complete ~~~')  
        
    @classmethod
    def tearDownClass(cls):
        print('!!! Class teardown for Test(Test_base_functions) !!!\n\n\n')


#unittest.main(argv=[''], verbosity=2, exit=False)


#!/usr/bin/env python
# coding: utf-8

# # BASE FUNCTIONS
# ### Functions for:
# - minimum down payment
# - mortgage rate by term
# - mortgage insurance
# - monthly payments
# - optimal monthly payment
# - total interest

# In[1]:

import numpy as np

from mortgage_filter.exceptions import *


def min_downpayment(price):
    """Returns the minimum downpayment required for a real estate
    price defined by the Financial Consumer Agency of Canada.
    (https://www.canada.ca/en/financial-consumer-agency/services/mortgages/down-payment.html)

    Arguments
    ----------
    price : numeric
        Property price or avereage area property price

    Return
    ------
    float
        minimum downpayment
    """
    try:
        if price < 0:
            print("Invalid price")
            return None
        elif price < 500000:
            return price * 0.05
        elif price < 1000000:
            return 500000 * 0.05 + (price - 500000) * 0.1
        return price * 0.2

    except TypeError:
        print("Invalid price input. Must be of type numeric")
        return None


def mort_rate(term):
    """If no mortgage rate is specified this function can be used to
    return an estimated mortgage rate based on a regression fit (R^2 = 0.926)
    on average Canadian mortgage rates for possible term lengths.
    (https://www.superbrokers.ca/tools/mortgage-rates-comparison/)

    Arguments
    ----------
    term : int
        contract length in years (from 1 to 10 years)

    Return
    ------
    float
        interest rate
    """
    try:
        x = term
        if x < 1:
            raise TermError
        elif x > 10:
            raise TermError(
                "Lengths greater than 10 years are not typically available."
            )
        elif isinstance(x, float):
            print(
                "Warning: Term lengths are typically in whole years not fractions of years."
            )
        return round((0.0167 * x**2 - 0.0337 * x + 1.6851), 3)

    except TermError as TE:
        print(
            f"{TE} \nTerms must range from 1 to 10 years, but calculation will be performed anyway."
        )
        return round((0.0167 * x**2 - 0.0337 * x + 1.6851), 3)


def mortgage_insurance(price, downpayment):
    """Returns the cost of mortgage insurance.

    Insurance rates are calculated from loan to asset price ratio.
    Rates are applied to the loan to generate a lump sum amount that's
    then added to the principal of the loan to give mortgage insurance.

    Arguments
    ----------
    price : numeric
        Property price

    downpayment : int or float
        Downpayment on property

    Return
    ------
    float
        Mortgage insurance
    """
    try:
        DP_proportion = downpayment / price

        # if downpayment more than 20% of the house price no mortgage insurance required.
        if DP_proportion >= 0.2:
            return 0
        elif DP_proportion < 0.05:
            raise MinimalDPError("Downpayment must be at least 5% the asset value")

        loan_to_price = (price - downpayment) / price
        x = loan_to_price

        # loan to price ratio determines insurance rate
        insurance_rate = (
            2924.5 * x**4 - 9340.3 * x**3 + 11116 * x**2 - 5830.8 * x + 1137.1
        ) / 100

        # mortgage insurance is a % applied to the mortgage amount (price - downpayment)
        return round(((price - downpayment) * insurance_rate), 2)

    except TypeError:
        print("Bad entry type. Received:", type(price), type(downpayment))
        return None
    except MinimalDPError as PE:
        print(PE, "\nInput value is too low to be legally considered.")
        return None
    except ZeroDivisionError:
        print("Price cannot be zero.")
        return None


def monthly_payment(principal, mortgage_rate, amortization, months=False):
    """Returns the monthly payment required to meet the given amortization period.
    Assumes payments occur on a monthly basis.

    Arguments
    ----------
    principal : numeric

    mortgage_rate : float
        Annual mortgage rate (loan interest)

    amortization: int
        Amortization period in years (or in months if months == True)

    months : bool
        (Optional) if True, amortization period is interpreted in months (default = False)

    Return
    ------
    float
        monthly payment
    """

    R = mortgage_rate / 100 / 12 + 1  ## monthly interest rate

    if months == True:
        n = amortization  ## if specified in months, amortization = the number of payments
    else:
        n = (
            amortization * 12
        )  ## convert amortization in years to the number of monthly payments

    monthly_contribution = principal * ((R**n) * (1 - R) / (1 - R**n))

    return round(monthly_contribution, 2)


def optimal_monthly_payment(principal, mortgage_rate, max_monthly_payment):
    """Returns the first amortization period which has a monthly payment
    less than your max_monthly_payment (ie. within budget). The shortest
    possible amortization period has the lowest long term interest cost.

    Arguments
    ----------
    principal : numeric

    mortgage_rate : float
          Annual mortgage rate (loan interest)

    max_monthly_payment: numeric
        Your max affordable monthly contribution

    Return
    ------
    list
        mp: monthly payment for a given amortization
        i: amortization period in years
    """
    try:
        for i in range(1, 26):
            mp = monthly_payment(principal, mortgage_rate, i, months=False)
            if mp <= max_monthly_payment:
                return [mp, i]
        return [np.nan, np.nan]

    except TypeError:
        print(
            "Bad entry type. Received:",
            type(principal),
            type(mortgage_rate),
            type(max_monthly_payment),
        )
        return None


def total_interest(principal, mortgage_rate, monthly_payment):
    """Returns the cumulative interest paid on a given principal, mortgage rate, and monthly payment.

    Arguments
    ----------
    principal : numeric

    mortgage_rate : float
        Annual mortgage rate (loan interest)

    amortization: int
        Amortization period in years (or in months if months == True)

    monthly_payment : bool
        Monthly contribution towards the principal

    Return
    ------
    float
        Cumulative interest paid
    """
    try:
        R = mortgage_rate / 1200  ## monthly interest rate
        CumInterest = 0

        i = principal * R
        new_p = principal + i - monthly_payment

        while new_p > 0:
            CumInterest += i
            i = new_p * R
            new_p = new_p + i - monthly_payment

            if new_p >= new_p - i + monthly_payment:
                print(
                    "Monthly contribution is insufficient to pay off the original Principal."
                )
                return None

        return round(CumInterest, 2)

    except TypeError:
        print(
            "Bad entry type. Received:",
            type(principal),
            type(mortgage_rate),
            type(monthly_payment),
        )
        return None



import pathlib

import setuptools

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setuptools.setup(
    name="mortgage-pkg",
    version="1.0.1",
    packages=setuptools.find_packages(exclude=["tests*"]),
    license="MIT",
    description="A package for filtering real estate opportunities based on your financial situation",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/lukavuko/mortgage-filter-package",
    author="Luka Vukovic",
    author_email="luka.vuko@outlook.com",
    keywords="mortgage, real estate, property filter, affordability, finance, calculator",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Developers",
        "Topic :: Office/Business :: Financial",
        "Topic :: Scientific/Engineering :: Mathematics",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    include_package_data=True,
    install_requires=["pandas", "numpy"],
)
